__copyright__ = "Copyright (C) 2009 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

CLUDA_PREAMBLE = """
#define local_barrier() barrier(CLK_LOCAL_MEM_FENCE);

#define WITHIN_KERNEL /* empty */
#define KERNEL __kernel
#define GLOBAL_MEM __global
#define LOCAL_MEM __local
#define LOCAL_MEM_ARG __local
#define REQD_WG_SIZE(X,Y,Z) __attribute__((reqd_work_group_size(X, Y, Z)))

#define LID_0 ((ptrdiff_t) get_local_id(0))
#define LID_1 ((ptrdiff_t) get_local_id(1))
#define LID_2 ((ptrdiff_t) get_local_id(2))

#define GID_0 ((ptrdiff_t) get_group_id(0))
#define GID_1 ((ptrdiff_t) get_group_id(1))
#define GID_2 ((ptrdiff_t) get_group_id(2))

#define LDIM_0 ((ptrdiff_t) get_local_size(0))
#define LDIM_1 ((ptrdiff_t) get_local_size(1))
#define LDIM_2 ((ptrdiff_t) get_local_size(2))

#define GDIM_0 ((ptrdiff_t) get_num_groups(0))
#define GDIM_1 ((ptrdiff_t) get_num_groups(1))
#define GDIM_2 ((ptrdiff_t) get_num_groups(2))

% if double_support:
    #if __OPENCL_C_VERSION__ < 120
    #pragma OPENCL EXTENSION cl_khr_fp64: enable
    #endif
% endif
"""
